# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""This module provides utilities to display quantum-viz from a Python script."""
import json
import tempfile
import warnings
import webbrowser
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Optional
from typing import TYPE_CHECKING
from typing import Union

from .widget import DEFAULT_STYLE
from .widget import Style

if TYPE_CHECKING:
    from qiskit import QuantumCircuit

HTML_TEMPLATE = """<!DOCTYPE html>
<html lang="en">
<head>
  <meta charset="UTF-8">
  <meta name="viewport" content="width=device-width, initial-scale=1.0">
  <title>quantum-viz.js visualization</title>
</head>
<body>
  <div id="msg" style="font-size: large; font-weight: bolder;">
    There was some error with the script
  </div>
  <div id="circuit_div"></div>
  <script src="https://unpkg.com/@microsoft/quantum-viz.js{0}"></script>
  <script type="text/javascript">
    var circuit = {1};
    if (typeof qviz != 'undefined' && typeof circuit != undefined) {{
      document.getElementById('msg').style.display = 'none';
      var displayDiv = document.getElementById('circuit_div');
      if (displayDiv != null) {{
        qviz.draw(circuit, displayDiv, qviz.STYLES['{2}']);
      }}
    }}
  </script>

</body>
</html>
"""
SUFFIX = "_qviz.html"

STYLES = list(Style)


class UnsupportedStyleWarning(UserWarning):
    """A warning raised when an unsupported style is chosen."""


def _create_file(
    circuit: Union[Dict[str, Any], "QuantumCircuit"],
    filename: Union[str, Path, None] = None,
    style: Style = DEFAULT_STYLE,
    version: Optional[str] = None,
    **kwargs,
) -> Path:
    if filename is None:
        file = tempfile.NamedTemporaryFile(suffix=SUFFIX, delete=False)
        file.close()
        path = Path(file.name)
    else:
        path = Path(filename)

    if style not in STYLES:
        warnings.warn(
            f"The selected style '{style}' is not supported and will be ignored.\n"
            f"The supported styles are {STYLES}",
            UnsupportedStyleWarning,
        )

    if version is None:
        version = ""  # Use the latest version
    else:
        version = "@" + version

    if not isinstance(circuit, dict):
        from .qiskit_parser import qiskit2dict

        circuit = qiskit2dict(circuit, **kwargs)

    qviz_json = json.dumps(circuit)

    html = HTML_TEMPLATE.format(version, qviz_json, style)
    path.write_text(html)
    return path


def display(
    circuit: Union[Dict[str, Any], "QuantumCircuit"],
    filename: Union[str, Path, None] = None,
    style: Style = DEFAULT_STYLE,
    version: Optional[str] = None,
    **kwargs,
) -> None:
    """Render the given circuit using quantum-viz on a new browser window."""
    path = _create_file(circuit, filename, style, version, **kwargs)
    webbrowser.open(f"file://{path.absolute()}")



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""quantum-viz Viewer is a Jupyter Widget that displays the quantum-viz.js circuit
visualizer.
"""  # noqa: D400, D205
import json
import uuid
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from IPython.display import display
from IPython.display import HTML
from varname import varname
from varname.utils import ImproperUseError

# The default quantum-viz.js version to use.
VERSION = "1.0.2"
# Rel file path for Javascript source
JS_SOURCE = "qviz.min"
BASE_URL = "https://unpkg.com/@microsoft/quantum-viz.js{version_suffix}/dist"


class Style(str, Enum):
    """Supported styles for the quantum-viz.js widget."""

    DEFAULT = "Default"
    BNW = "BlackAndWhite"
    INV = "Inverted"


DEFAULT_STYLE = Style.DEFAULT

_HTML_STR_FORMAT = """
<script type="text/javascript">
require.config({{
    paths: {{
        qviz: '{base_url}/{js_source}'
    }}
}});
require(['qviz'], function(qviz) {{
    const circuit = {data};
    const targetDiv = document.getElementById('JSApp_{uid}');
    if (targetDiv != null) {{
        qviz.draw(circuit, targetDiv, qviz.STYLES['{style}']);
    }}
}});
</script>
<div id="JSApp_{uid}"></div>
<div id="msg"></div>
"""


class Viewer:
    """Jupyter widget for displaying Quantum-viz quantum circuit."""

    n = 0

    def __init__(
        self,
        circuit: Dict[str, Any],
        version: Optional[str] = VERSION,
        style: Style = DEFAULT_STYLE,
        width: int = 400,
        height: int = 350,
    ):
        """
        Create Viewer instance.

        :param circuit: Quantum circuit
        :type circuit: dict
        :param width: Widget width in pixels, defaults to 400
        :type width: int, optional
        :param height: Widget height in pixels, defaults to 350
        :type height: int, optional
        """
        try:
            self.name = varname()
        except ImproperUseError:
            self.name = "_"

        if not isinstance(circuit, dict):
            from .qiskit_parser import qiskit2dict, QuantumCircuit

            if not isinstance(circuit, QuantumCircuit):
                raise TypeError(
                    f"Received a circuit of an unsupported type: {type(circuit)}"
                )
            circuit = qiskit2dict(circuit)

        self.width = width
        self.height = height
        self.value = json.dumps(circuit)
        self.base_url = self._get_base_url(version)
        self.style = style
        self._uids: List[str] = []

    @staticmethod
    def _get_base_url(version: Optional[str]) -> str:
        if version is None:
            version_suffix = ""  # Use the latest version
        else:
            version_suffix = f"@{version}"
        return BASE_URL.format(version_suffix=version_suffix)

    def _gen_uid(self) -> str:
        """Generate unique identifier for javascript applet. Returns UID."""
        uid = str(uuid.uuid1()).replace("-", "")
        # Keep track of all UIDs
        self._uids.append(uid)
        return uid

    def html_str(self, uid: Optional[str] = None) -> str:
        """Return an HTML string that contains the widget.

        :param uid: Unique identifier of widget
        :type uid: str or None
        :return: HTML string for displaying widget
        :rtype: str
        """
        if uid is None:
            uid = self._gen_uid()
        Viewer.n += 1
        return _HTML_STR_FORMAT.format(
            base_url=self.base_url,
            js_source=JS_SOURCE,
            uid=uid,
            data=self.value,
            style=self.style,
        )

    def _ipython_display_(self) -> None:
        """Display the widget with IPython."""
        viewer = HTML(self.html_str())
        display(viewer)

    def browser_display(self) -> None:
        """Display the widget in the browser."""
        raise NotImplementedError



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""This module provides methods to serialize a Qiskit circuit into a qviz dictionary."""  # noqa: B950
from enum import IntEnum
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

try:
    import qiskit  # noqa: F401
except ImportError as e:
    raise ImportError(
        '`qiskit` was not found, try to `pip install "quantum-viz[qiskit]"`'
    ) from e

from qiskit.circuit import QuantumCircuit, Qubit, Clbit, ClassicalRegister
from qiskit.circuit.instruction import Instruction
from qiskit.circuit.controlledgate import ControlledGate
from qiskit.circuit.measure import Measure
from qiskit.circuit.barrier import Barrier
from qiskit.circuit.reset import Reset
from qiskit.circuit.parameter import Parameter
from qiskit.circuit.library import (
    IGate,
    XGate,
    YGate,
    ZGate,
    HGate,
    UGate,
    U1Gate,
    U2Gate,
    U3Gate,
    PhaseGate,
    RGate,
    SwapGate,
    RXGate,
    RYGate,
    RZGate,
    RXXGate,
    RYYGate,
    RZXGate,
    RZZGate,
    iSwapGate,
    DCXGate,
    SGate,
    SdgGate,
    TGate,
    TdgGate,
    SXGate,
    SXdgGate,
)
from qiskit.circuit.quantumcircuitdata import QuantumCircuitData

INSTRUCTION_TYPE_TO_NAME: Dict[Type[Instruction], str] = {
    IGate: "I",
    XGate: "X",
    YGate: "Y",
    ZGate: "Z",
    HGate: "H",
    UGate: "U",
    U1Gate: "U1",
    U2Gate: "U2",
    U3Gate: "U3",
    PhaseGate: "P",
    RGate: "R",
    SwapGate: "SWAP",
    RXGate: "Rx",
    RYGate: "Ry",
    RZGate: "Rz",
    RXXGate: "Rxx",
    RYYGate: "Ryy",
    RZXGate: "Rzx",
    RZZGate: "Rzz",
    iSwapGate: "Iswap",
    DCXGate: "Dcx",
    SGate: "S",
    SdgGate: "Sdg",
    TGate: "T",
    TdgGate: "Tdg",
    SXGate: "√X",
    SXdgGate: "√Xdg",
    Measure: "measure",
}

X_GATE_NAME = INSTRUCTION_TYPE_TO_NAME[XGate]
MEASURE_NAME = INSTRUCTION_TYPE_TO_NAME[Measure]


class RegisterType(IntEnum):  # noqa: D101
    QUBIT = 0
    CLASSICAL = 1


class ConditionalRender(IntEnum):  # noqa: D101
    ALWAYS = 0
    ON_ZERO = 1
    ON_ONE = 2
    AS_GROUP = 3


def qiskit2dict(circ: QuantumCircuit, **kwargs) -> Dict[str, List]:
    """Convert a Qiskit circuit to a qviz dictionary."""
    return QiskitCircuitParser(circ, **kwargs).qviz_dict


class QiskitCircuitParser:
    """This class parses a Qiskit circuit into a qviz dictionary."""

    QUBITS_KEY = "qubits"
    OPERATIONS_KEY = "operations"

    def __init__(
        self,
        circuit: QuantumCircuit,
        precision: int = 2,
        max_recursion_depth: Optional[int] = None,
        skip_barriers: bool = True,
    ) -> None:
        """
        Create a QiskitCircuitParser object.

        :param circuit: qiskit quantum circuit to be parsed
        :param precision: the decimal precision of float gate parameters to display
        :param max_recursion_depth: the maximal recursion depth to parse, if None -
          parse until the basis gates are reached
        :param skip_barriers: whether to omit barriers in the output or not
        """
        self.qc = circuit
        self.precision = precision
        self.max_recursion_depth = max_recursion_depth
        self.skip_barriers = skip_barriers
        self.qviz_dict: Dict[str, List] = {
            self.QUBITS_KEY: [],
            self.OPERATIONS_KEY: [],
        }
        self.qubit2id: Dict[Qubit, int] = dict()
        self._init_qubits()
        self._clbit2id: Dict[Clbit, Tuple[int, int]] = dict()
        self._update_qviz_dict()

    def _init_qubits(self) -> None:
        qubits_range = range(self.qc.num_qubits)
        self.qubit2id = dict(zip(self.qc.qubits, qubits_range))
        self.qubits = [{"id": i} for i in qubits_range]

    @property
    def qubits(self) -> List[Dict[str, int]]:
        """Get the list of qubits used by the circuit."""
        return self.qviz_dict[self.QUBITS_KEY]
        return self.qviz_dict[self.QUBITS_KEY]

    @qubits.setter
    def qubits(self, value: List[Dict[str, int]]) -> None:
        self.qviz_dict[self.QUBITS_KEY] = value

    @property
    def operations(self) -> List[Dict]:
        """Get the list of operations used by the circuit."""
        return self.qviz_dict[self.OPERATIONS_KEY]

    @operations.setter
    def operations(self, value: List[Dict]) -> None:
        self.qviz_dict[self.OPERATIONS_KEY] = value

    def _get_qubit_def(self, qubit: Qubit) -> Dict[str, int]:
        return {"qId": self.qubit2id[qubit]}

    def _get_qubit_list_def(self, qubits: List[Qubit]) -> List[Dict[str, int]]:
        return [self._get_qubit_def(qubit) for qubit in qubits]

    def _get_clbit_def(
        self, clbit: Clbit, qubit: Optional[Qubit] = None
    ) -> Dict[str, int]:
        if clbit in self._clbit2id:
            q_id, c_id = self._clbit2id[clbit]
        else:
            if qubit is None:
                raise NotImplementedError(
                    "A classical bit has to defined first as a measurement result of a "
                    "qubit"
                )
            q_id = self.qubit2id[qubit]
            c_id = self.qubits[q_id].get("numChildren", 0)
            self.qubits[q_id]["numChildren"] = c_id + 1
            self._clbit2id[clbit] = (q_id, c_id)
        return {"type": RegisterType.CLASSICAL.value, "qId": q_id, "cId": c_id}

    def _update_qviz_dict(self) -> None:
        qc = self.qc
        self.operations += [
            {
                "gate": qc.name,
                "targets": self._get_qubit_list_def(qc.qubits),
            }
        ]
        for instruction, qargs, cargs in self._filter_circuit_data(qc.data):
            self.operations[0]["children"] = self.operations[0].get("children", []) + [
                self._parse_operation(instruction, qargs, cargs, depth=1)
            ]

    def _filter_circuit_data(self, data: QuantumCircuitData) -> Iterator:
        if self.skip_barriers:
            # check whether the instruction is not a barrier
            return filter(lambda elem: not isinstance(elem[0], Barrier), data)
        return data

    def _is_recursion_depth_exceeded(self, depth: int) -> bool:
        return self.max_recursion_depth is not None and depth > self.max_recursion_depth

    def _parse_operation(
        self,
        instruction: Instruction,
        qargs: List[Qubit],
        cargs: List[Clbit],
        depth: int,
    ) -> Dict:

        op_dict: Dict = {"gate": self._get_instruction_name(instruction)}

        if instruction.params:
            self._add_params(op_dict, instruction)

        if isinstance(instruction, Reset):
            self._add_reset(op_dict, qubit=qargs[0], depth=depth)
        elif isinstance(instruction, Measure):
            self._add_measure(op_dict, qargs, cargs)
        elif isinstance(instruction, ControlledGate):
            # a controlled gate may change the gate name (i.e. CX -> X)
            self._add_controlled_gate(op_dict, instruction, qargs)
        else:
            op_dict["targets"] = self._get_qubit_list_def(qargs)

        self._add_children(op_dict, instruction, qargs, cargs, depth)

        if instruction.condition:
            # _update_condition must update op_dict last (after any other method)
            op_dict = self._update_condition(op_dict, instruction)

        return op_dict

    def _add_children(
        self,
        op_dict: Dict,
        instruction: Instruction,
        qargs: List[Qubit],
        cargs: List[Clbit],
        depth: int,
    ) -> None:
        if (instruction.definition is not None) and (
            not self._is_recursion_depth_exceeded(depth + 1)
        ):
            sub_circuit: QuantumCircuit = instruction.definition
            # Since the `index` property of bits is deprecated - create mappers between
            # the bits and their indices in the circuit.
            # Use the builtin `QuantumCircuit.find_bit` method when it is released
            qubits_mapper = dict(zip(sub_circuit.qubits, range(sub_circuit.num_qubits)))
            clbits_mapper = dict(zip(sub_circuit.clbits, range(sub_circuit.num_clbits)))
            for sub_instruction, sub_qargs, sub_cargs in self._filter_circuit_data(
                sub_circuit.data
            ):
                # Update the args to those of the containing circuit
                sub_qargs = [qargs[qubits_mapper[qubit]] for qubit in sub_qargs]
                sub_cargs = [cargs[clbits_mapper[clbit]] for clbit in sub_cargs]
                op_dict["children"] = op_dict.get("children", []) + [
                    self._parse_operation(
                        sub_instruction, sub_qargs, sub_cargs, depth=depth + 1
                    )
                ]

    def _add_controlled_gate(
        self, op_dict: Dict, cgate: ControlledGate, qargs: List[Qubit]
    ) -> None:
        ctrl_state = cgate.ctrl_state
        if not all([int(c) for c in str(ctrl_state)]):
            raise NotImplementedError(
                f"The controlled gate {cgate} is controlled by a state that "
                f"is not all 1's: {ctrl_state}"
            )
        num_ctrl_qubits = cgate.num_ctrl_qubits
        ctrl_qubits = qargs[:num_ctrl_qubits]
        target_qubits = qargs[num_ctrl_qubits:]
        op_dict["isControlled"] = True
        op_dict["gate"] = self._get_instruction_name(cgate.base_gate)
        op_dict["controls"] = self._get_qubit_list_def(ctrl_qubits)
        op_dict["targets"] = self._get_qubit_list_def(target_qubits)

    def _add_measure(
        self, op_dict: Dict, qargs: List[Qubit], cargs: List[Clbit]
    ) -> None:
        if len(qargs) != 1 or len(cargs) != 1:
            raise ValueError
        qubit = qargs[0]
        clbit = cargs[0]
        op_dict["isMeasurement"] = True
        op_dict["controls"] = [self._get_qubit_def(qubit)]
        op_dict["targets"] = [self._get_clbit_def(clbit, qubit)]

    def _format_param(self, param: Union[int, float, complex, Parameter]) -> str:
        if isinstance(param, (float, complex)):
            return f"{param:.{self.precision}f}"
        return str(param)

    def _add_params(self, op_dict: Dict, instruction: Instruction) -> None:
        params_map = map(self._format_param, instruction.params)
        op_dict["displayArgs"] = f"({', '.join(params_map)})"

    @staticmethod
    def _get_instruction_name(instruction: Instruction) -> str:
        return INSTRUCTION_TYPE_TO_NAME.get(type(instruction), instruction.name)

    def _add_reset(self, op_dict: Dict, qubit: Qubit, depth: int) -> None:
        """Reset logic - measure and apply X gate if the measurement yields 1."""
        qubit_def = self._get_qubit_def(qubit)
        op_dict["targets"] = [qubit_def]
        clbit_def = self._get_clbit_def(Clbit(), qubit)  # create a new classical bit

        if not self._is_recursion_depth_exceeded(depth + 1):
            # Add a simple logic for the reset instruction
            op_dict["children"] = [
                {
                    "gate": MEASURE_NAME,
                    "isMeasurement": True,
                    "controls": [qubit_def],
                    "targets": [clbit_def],
                },
                {
                    "gate": "Conditional",
                    "isConditional": True,
                    "controls": [clbit_def],
                    "targets": [],
                    "children": [
                        {
                            "gate": X_GATE_NAME,
                            "targets": [qubit_def],
                            "conditionalRender": ConditionalRender.ON_ONE.value,
                        }
                    ],
                },
            ]

    def _update_condition(self, op_dict: Dict, instruction: Instruction) -> Dict:
        classical, val = instruction.condition
        if isinstance(classical, ClassicalRegister) and classical.size > 1:
            raise NotImplementedError(
                "quantum-viz.js does not support conditioning an instruction on "
                "more than a single classical bit"
            )
        if not (isinstance(val, bool) or val in (0, 1)):
            raise ValueError(
                "A single classical bit can hold only 0, 1 or a boolean, but the "
                f"condition value is {val}"
            )
        if isinstance(classical, ClassicalRegister):
            clbit = Clbit(classical, index=0)
        else:
            clbit = classical

        if val:
            render_condition = ConditionalRender.ON_ONE.value
        else:
            render_condition = ConditionalRender.ON_ZERO.value

        op_dict["conditionalRender"] = render_condition

        conditioned_op_dict = {
            "gate": "Conditional",
            "isConditional": True,
            "controls": [self._get_clbit_def(clbit)],
            "targets": [],
            "children": [op_dict],
        }

        return conditioned_op_dict



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['quantum_viz']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.1,<9.0.0', 'ipython>=8.0.3,<9.0.0', 'varname>=0.8.1,<0.9.0']

extras_require = \
{'qiskit': ['qiskit-terra>=0.19.1,<1']}

setup_kwargs = {
    'name': 'quantum-viz',
    'version': '1.0.5',
    'description': 'quantum-viz.js Python tools',
    'long_description': '# quantum-viz\n\n`quantum-viz` is the Python package companion of [quantum-viz.js](https://github.com/microsoft/quantum-viz.js), a JavaScript package that supports visualizing any arbitrary quantum gate, classical control logic and collapsed grouped blocks of gates using JSON-formatted input data. `quantum-viz` contains a Jupyter widget and will also include support for translating quantum circuits written in common quantum programming libraries to JSON using the `quantum-viz.js` JSON schema.\n\n![quantum-viz screenshot](https://user-images.githubusercontent.com/4041805/137234877-6a529652-a3b9-48c6-9d3c-c2b9d1e11855.gif)\n\n## Installation\n\nYou can install the *quantum-viz.js widget* via `pip` from PyPI:\n\n```bash\npip install quantum-viz\n```\n\n## Example\n\nTo use the quantum-viz widget, run the below example code in a [Jupyter notebook](https://jupyter.org/) cell:\n\n```python\nfrom quantum_viz import Viewer\n\n# Create a quantum circuit that prepares a Bell state\ncircuit = {\n    "qubits": [{ "id": 0 }, { "id": 1, "numChildren": 1 }],\n    "operations": [\n        {\n            "gate": \'H\',\n            "targets": [{ "qId": 0 }],\n        },\n        {\n            "gate": \'X\',\n            "isControlled": "True",\n            "controls": [{ "qId": 0 }],\n            "targets": [{ "qId": 1 }],\n        },\n        {\n            "gate": \'Measure\',\n            "isMeasurement": "True",\n            "controls": [{ "qId": 1 }],\n            "targets": [{ "type": 1, "qId": 1, "cId": 0 }],\n        },\n    ],\n}\n\nwidget = Viewer(circuit)\nwidget # Display the widget\n```\n\n![quantum-viz example](https://user-images.githubusercontent.com/4041805/137230540-b523dc76-29c7-48e2-baa3-34d4ee0a17a1.PNG)\n\n## Qiskit Integration\n\nBy installing the optional `[qiskit]` dependency, you can leverage Qiskit\'s `QuantumCircuit` APIs\nto define the circuit and render it using the `Viewer` widget on Jupyter, for example:\n\n```python\nfrom qiskit import QuantumRegister, ClassicalRegister, QuantumCircuit\nfrom quantum_viz import Viewer\n\nqr = QuantumRegister(3, \'q\')\nanc = QuantumRegister(1, \'ancilla\')\ncr = ClassicalRegister(3, \'c\')\nqc = QuantumCircuit(qr, anc, cr)\n\n\nqc.h(qr[0:3])\nqc.x(anc[0])\nqc.h(anc[0])\nqc.cx(qr[0:3], anc[0])\nqc.h(qr[0:3])\nqc.barrier(qr)\nqc.measure(qr, cr)\n\nViewer(qc)\n```\n\nOptionally, you can also import the `display` method from `quantum_viz.utils` to render the circuit on a new browser window:\n\n```python\nfrom quantum_viz.utils import display\ndisplay(qc)\n```\n\n## Contributing\n\nCheck out our [contributing guidelines](https://github.com/microsoft/quantum-viz.js/blob/main/quantum-viz/CONTRIBUTING.md) to find out how you can contribute to quantum-viz.\n',
    'author': 'Microsoft Corporation',
    'author_email': 'que-contacts@microsoft.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/microsoft/quantum-viz.js',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
